/*
 * rsaglue1.c - These functions wrap and unwrap message digests (MDs) and
 * data encryption keys (DEKs) in padding and RSA-encrypt them into
 * multi-precision integers.  This layer of abstraction was introduced
 * to allow the transparent use of either the RSAREF Cryptographic
 * Toolkit from RSA Data Security Inc for the RSA calculations
 * or, Philip Zimmermann's mpi library for the
 * RSA calculations.  The rsaglue.c module from PGP version 2.3a performs
 * the same functions as this module, but can be compiled to select the
 * use of mpilib functions instead of RSAREF as the underlying math engine.
 *
 * This file uses MPILIB to perform the actual encryption and decryption.
 * It uses the same PKCS format as RSAREF, although it also accepts an older
 * format used in PGP 2.1.
 *
 * (c) Copyright 1990-1996 by Philip Zimmermann.  All rights reserved.
 * The author assumes no liability for damages resulting from the use
 * of this software, even if the damage results from defects in this
 * software.  No warranty is expressed or implied.
 *
 * Note that while most PGP source modules bear Philip Zimmermann's
 * copyright notice, many of them have been revised or entirely written
 * by contributors who frequently failed to put their names in their
 * code.  Code that has been incorporated into PGP from other authors
 * was either originally published in the public domain or is used with
 * permission from the various authors.
 *
 * PGP is available for free to the public under certain restrictions.
 * See the PGP User's Guide (included in the release package) for
 * important information about licensing, patent restrictions on
 * certain algorithms, trademarks, copyrights, and export controls.
 */

#include <string.h>     /* for mem*() */
#include <stdio.h>
#include "xnew.h"
#include "mpilib.h"
#include "mpiio.h"
#include "pgp.h"
#include "language.h"
#include "rsaglue.h"
#include "random.h" /* for cryptRandByte() */

/* No RSADSI credit for MPI version */
char signon_legalese[] = "";

/* These functions hide all the internal details of RSA-encrypted
 * keys and digests.  They owe a lot of their heritage to
 * the preblock() and postunblock() routines in mpiio.c.
 */

#include "asn.h"

int
rsa_public_encrypt(unitptr outbuf, byteptr inbuf, short bytes,
     unitptr E, unitptr N)
/* Encrypt a DEK with a public key.  Returns 0 on success.
 * <0 means there was an error.
 * -1: Generic error
 * -3: Key too big
 * -4: Key too small
 */
{
    XNEW(temp,unit,MAX_UNIT_PRECISION);
    unsigned int blocksize;
    int i;  /* Temporary, and holds error codes */
    byte *p = (byte *)temp;
    int Xretval = 0;
#define XRETURN(x) {Xretval = x; goto exit_rsa_public_encrypt;}

    /*
     * We are building the mpi in place, except for a possible
     * byte-order swap to little-endian at the end.  Thus, we
     * need to fill the buffer with leading 0's in the unused
     * most significant byte positions.
     */
    blocksize = countbytes(N) - 1;  /* Bytes available for user data */
    for (i = units2bytes(global_precision) - blocksize; i > 0; --i)
        *p++ = 0;
    /*
     * Both the PKCS and PGP 2.0 key formats add a type byte, and a
     * a framing byte of 0 to the user data.  The remaining space
     * is filled with random padding.  (PKCS requires that there be
     * at least 1 byte of padding.)
     */
    i = blocksize - 2 - bytes;

    if (i < 1)      /* Less than minimum padding? */
        XRETURN (-4);
    *p++ = CK_ENCRYPTED_BYTE;   /* Type byte */
    while (i)           /* Non-zero random padding */
        if ((*p = cryptRandByte()))
            ++p, --i;
    *p++ = 0;           /* Framing byte */
    memcpy(p, inbuf, bytes);    /* User data */

    mp_convert_order((byte *)temp);     /* Convert buffer to MPI */
    i = mp_modexp(outbuf, temp, E, N);  /* Do the encryption */
    if (i < 0)
        i = -1;

/*Cleanup:*/
    mp_burn(temp);
    XRETURN (i < 0 ? i : 0);

#undef XRETURN
    exit_rsa_public_encrypt:
    XDEL(temp);
    return Xretval;
} /* rsa_public_encrypt */

int
rsa_private_encrypt(unitptr outbuf, byteptr inbuf, short bytes,
     unitptr E, unitptr D, unitptr P, unitptr Q, unitptr U, unitptr N, int Hash)
/* Encrypt a message digest with a private key.
 * Returns <0 on error:
 * -1: generic error
 * -3: Key too big
 * -4: Key too small
 */
{
    XNEW(temp,unit,MAX_UNIT_PRECISION);
    XNEW(DP,unit,MAX_UNIT_PRECISION); XNEW(DQ,unit,MAX_UNIT_PRECISION);
    byte *p;
    int i;
    unsigned int blocksize;

    mp_mult (temp, P, Q);
    if (mp_compare (temp, N)) {
	    fprintf(pgpout, LANG("\n\007Damaged private key. Private key files modified?\n"));
	    i = -1;
	    goto Cleanup;
	}

    /* PGP doesn't store these coefficents, so we need to compute them. */
    mp_move(temp,P);
    mp_dec(temp);
    mp_mod(DP,D,temp);
    mp_move(temp,Q);
    mp_dec(temp);
    mp_mod(DQ,D,temp);

    p = (byte *)temp;


    /* We are building the mpi in place, except for a possible
     * byte-order swap to little-endian at the end.  Thus, we
     * need to fill the buffer with leading 0's in the unused
     * most significant byte positions.
     */
    blocksize = countbytes(N) - 1;  /* Space available for data */
    for (i = units2bytes(global_precision) - blocksize; i > 0; --i)
        *p++ = 0;

    i = blocksize - 2 - bytes;      /* Padding needed */
    if (bytes == 20) {
        if (Hash == SHA1_ALGORITHM_BYTE)
            i -= sizeof(sha1_asn_array);        /* Space for type encoding */
        else if (Hash == RIPEMD160_ALGORITHM_BYTE)
            i -= sizeof(ripemd160_asn_array);        /* Space for type encoding */
        else if (Hash == HAVAL_5_160_ALGORITHM_BYTE)
            i -= sizeof(haval_5_160_asn_array);        /* Space for type encoding */
    } else if (bytes == 32) {
        if (Hash == SHA256_ALGORITHM_BYTE)
            i -= sizeof(sha256_asn_array);        /* Space for type encoding */
        else if (Hash == HAVAL_5_256_ALGORITHM_BYTE)
            i -= sizeof(haval_5_256_asn_array);        /* Space for type encoding */
    } else if (bytes == 48) {
        if (Hash == SHA384_ALGORITHM_BYTE)
            i -= sizeof(sha384_asn_array);        /* Space for type encoding */
    } else if (bytes == 64) {
        if (Hash == SHA512_ALGORITHM_BYTE)
            i -= sizeof(sha512_asn_array);        /* Space for type encoding */
    } else if (bytes == 24) {
        if (Hash == TIGER192_ALGORITHM_BYTE)
            i -= sizeof(tiger192_asn_array);        /* Space for type encoding */
    } else if (bytes == 40) {
        if (Hash == SHA1X_ALGORITHM_BYTE)
            i -= sizeof(sha1x_asn_array);        /* Space for type encoding */
    } else if (bytes == 16) {
        /* if (Hash == MD5_ALGORITHM_BYTE) */
            i -= sizeof(asn_array);     /* Space for type encoding */
    }
    if (i < 0) {
        i = -4;         /* Error code */
        goto Cleanup;
    }
    *p++ = MD_ENCRYPTED_BYTE;   /* Type byte */
    memset(p, ~0, i);       /* All 1's padding */
    p += i;
    *p++ = 0;           /* Zero framing byte */
    if (bytes == 20) {
        if (Hash == SHA1_ALGORITHM_BYTE) {
            memcpy(p, sha1_asn_array, sizeof(sha1_asn_array)); /* ASN data */
            p += sizeof(sha1_asn_array);
        } else if (Hash == RIPEMD160_ALGORITHM_BYTE) {
            memcpy(p, ripemd160_asn_array, sizeof(ripemd160_asn_array)); /* ASN data */
            p += sizeof(ripemd160_asn_array);
        } else if (Hash == HAVAL_5_160_ALGORITHM_BYTE) {
            memcpy(p, haval_5_160_asn_array, sizeof(haval_5_160_asn_array)); /* ASN data */
            p += sizeof(haval_5_160_asn_array);
        }
    } else if (bytes == 32) {
        if (Hash == SHA256_ALGORITHM_BYTE) {
            memcpy(p, sha256_asn_array, sizeof(sha256_asn_array)); /* ASN data */
            p += sizeof(sha256_asn_array);
        } else if (Hash == HAVAL_5_256_ALGORITHM_BYTE) {
            memcpy(p, haval_5_256_asn_array, sizeof(haval_5_160_asn_array)); /* ASN data */
            p += sizeof(haval_5_256_asn_array);
        }
    } else if (bytes == 48) {
        if (Hash == SHA384_ALGORITHM_BYTE) {
            memcpy(p, sha384_asn_array, sizeof(sha384_asn_array)); /* ASN data */
            p += sizeof(sha384_asn_array);
        }
    } else if (bytes == 64) {
        if (Hash == SHA512_ALGORITHM_BYTE) {
            memcpy(p, sha512_asn_array, sizeof(sha512_asn_array)); /* ASN data */
            p += sizeof(sha512_asn_array);
        }
    } else if (bytes == 24) {
        if (Hash == TIGER192_ALGORITHM_BYTE) {
            memcpy(p, tiger192_asn_array, sizeof(tiger192_asn_array)); /* ASN data */
            p += sizeof(tiger192_asn_array);
        }
    } else if (bytes == 40) {
        if (Hash == SHA1X_ALGORITHM_BYTE) {
            memcpy(p, sha1x_asn_array, sizeof(sha1x_asn_array)); /* ASN data */
            p += sizeof(sha1x_asn_array);
        }
    } else if (bytes == 16) {
        /* if (Hash == MD5_ALGORITHM_BYTE) { */
            memcpy(p, asn_array, sizeof(asn_array)); /* ASN data */
            p += sizeof(asn_array);
        /* } */
    }
    memcpy(p, inbuf, bytes);    /* User data */

    mp_convert_order((byte *)temp);
    i = mp_modexp_crt(outbuf, temp, P, Q, DP, DQ, U);   /* Encrypt */
    if (i < 0)
        i = -1;
    else if (mp_modexp (DP, outbuf, E, N) < 0 ||
		memcmp (DP, temp, blocksize)) {
	    fprintf(pgpout, LANG("\n\007Decryption check failed. Private key files modified?\n"));
	    i = -1;
	}

Cleanup:
    burn(temp);
    burn(DP);
    burn(DQ);

    XDEL(temp);
    XDEL(DP); XDEL(DQ);
    return i;
} /* rsa_private_encrypt */

/* Remove a signature packet from an MPI */
/* Thus, we expect constant padding and the MIC ASN sequence */
int
rsa_public_decrypt(byteptr outbuf, unitptr inbuf,
    unitptr E, unitptr N, int * Hash)
/* Decrypt a message digest using a public key.  Returns the number of bytes
 * extracted, or <0 on error.
 * -1: Corrupted packet.
 * -3: Key too big
 * -4: Key too small
 * -5: Maybe malformed RSA packet
 * -7: Unknown conventional algorithm
 * -9: Malformed RSA packet
 */
{
    XNEW(temp,unit,MAX_UNIT_PRECISION);
    unsigned int blocksize;
    int i;
    byte *front, *back;
    int Xretval = 0;
#define XRETURN(x) {Xretval = x; goto exit_rsa_public_decrypt;}

    if (Hash) *Hash = 0;
    i = mp_modexp(temp, inbuf, E, N);
    if (i < 0) {
        mp_burn(temp);
        XRETURN (-1);
    }
    mp_convert_order((byte *)temp);
    blocksize = countbytes(N) - 1;
    front = (byte *)temp;           /* Points to start of block */
    i = units2bytes(global_precision);
    back = front + i;           /* Points to end of block */
    i -= countbytes(N) - 1;         /* Expected leading 0's */

    /*
     * Strip off the padding.  This handles both PKCS and PGP 2.0
     * formats.  If we're using RSAREF2, we use the padding-removal
     * code in RSAPublicDecrypt, which accepts only PKCS style.
     * Oh, well.
     */

    if (i < 0)              /* This shouldn't happen */
        goto ErrorReturn;
    while (i--)             /* Extra bytes should be 0 */
        if (*front++)
            goto ErrorReturn;

    /* How to distinguish old PGP from PKCS formats.
     * The old PGP format ends in a trailing type byte, with
     * all 1's padding before that.  The PKCS format ends in
     * 16 bytes of message digest, preceded by an ASN string
     * which is not all 1's.
     */
    if (back[-1] == MD_ENCRYPTED_BYTE &&
        back[-17] == 0xff && back[-18] == 0xff) {
        /* Old PGP format: Padding is at the end */
        if (*--back != MD_ENCRYPTED_BYTE)
            goto ErrorReturn;
        if (*front++ != MD5_ALGORITHM_BYTE) {
            mp_burn(temp);
            XRETURN (-7);
        }
        while (*--back == 0xff) /* Skip constant padding */
            ;
        if (*back)      /* It should end with a zero */
            goto ErrorReturn;
        if (Hash) *Hash = MD5_ALGORITHM_BYTE;
    } else {
        /* PKCS format: padding at the beginning */
        if (*front++ != MD_ENCRYPTED_BYTE)
            goto ErrorReturn;
        while (*front++ == 0xff) /* Skip constant padding */
            ;
        if (front[-1])  /* First non-FF byte should be 0 */
            goto ErrorReturn;
        /* Then comes the ASN header */
        if (!memcmp(front, asn_array, sizeof(asn_array))) {
            front += sizeof(asn_array);
            if (Hash) *Hash = MD5_ALGORITHM_BYTE;
        } else if (!memcmp(front, sha1_asn_array, sizeof(sha1_asn_array))) {
            front += sizeof(sha1_asn_array);
            if (Hash) *Hash = SHA1_ALGORITHM_BYTE;
            if (back-front == 40)
                if (Hash) *Hash = SHA1X_ALGORITHM_BYTE;
        } else if (!memcmp(front, ripemd160_asn_array, sizeof(ripemd160_asn_array))) {
            front += sizeof(ripemd160_asn_array);
            if (Hash) *Hash = RIPEMD160_ALGORITHM_BYTE;
        } else if (!memcmp(front, sha256_asn_array, sizeof(sha256_asn_array))) {
            front += sizeof(sha256_asn_array);
            if (Hash) *Hash = SHA256_ALGORITHM_BYTE;
        } else if (!memcmp(front, sha384_asn_array, sizeof(sha384_asn_array))) {
            front += sizeof(sha384_asn_array);
            if (Hash) *Hash = SHA384_ALGORITHM_BYTE;
        } else if (!memcmp(front, sha512_asn_array, sizeof(sha512_asn_array))) {
            front += sizeof(sha512_asn_array);
            if (Hash) *Hash = SHA512_ALGORITHM_BYTE;
        } else if (!memcmp(front, tiger192_asn_array, sizeof(tiger192_asn_array))) {
            front += sizeof(tiger192_asn_array);
            if (Hash) *Hash = TIGER192_ALGORITHM_BYTE;
        } else if (!memcmp(front, haval_5_160_asn_array, sizeof(haval_5_160_asn_array))) {
            front += sizeof(haval_5_160_asn_array);
            if (Hash) *Hash = HAVAL_5_160_ALGORITHM_BYTE;
        } else if (!memcmp(front, haval_5_256_asn_array, sizeof(haval_5_256_asn_array))) {
            front += sizeof(haval_5_256_asn_array);
            if (Hash) *Hash = HAVAL_5_256_ALGORITHM_BYTE;
        //} else if (!memcmp(front, sha1x_asn_array, sizeof(sha1x_asn_array))) {
        //    front += sizeof(sha1x_asn_array);
        //    if (Hash) *Hash = SHA1X_ALGORITHM_BYTE;
        } else {
            mp_burn(temp);
            XRETURN (-7);
        }
    }

/* We're done - copy user data to outbuf */
    if (back < front)
        goto ErrorReturn;
    blocksize = back-front;
    memcpy(outbuf, front, blocksize);
    mp_burn(temp);
    XRETURN (blocksize);
ErrorReturn:
    mp_burn(temp);
    XRETURN (-9);

#undef XRETURN
    exit_rsa_public_decrypt:
    XDEL(temp);
    return Xretval;
} /* rsa_public_decrypt */

/* We expect to find random padding and an encryption key */
int
rsa_private_decrypt(byteptr outbuf, unitptr inbuf,
     unitptr E, unitptr D, unitptr P, unitptr Q, unitptr U, unitptr N)
/* Decrypt an encryption key using a private key.  Returns the number of bytes
 * extracted, or <0 on error.
 * -1: Generic error
 * -3: Key too big
 * -4: Key too small
 * -5: Maybe malformed RSA
 * -7: Unknown conventional algorithm
 * -9: Malformed RSA packet
 */
{
    byte *back;
    byte *front;
    unsigned int blocksize;
    XNEW(temp,unit,MAX_UNIT_PRECISION);
    XNEW(DP,unit,MAX_UNIT_PRECISION); XNEW(DQ,unit,MAX_UNIT_PRECISION);
    int i;
    int Xretval = 0;
#define XRETURN(x) {Xretval = x; goto exit_rsa_private_decrypt;}

    /* PGP doesn't store (d mod p-1) and (d mod q-1), so compute 'em */
    mp_move(temp,P);
    mp_dec(temp);
    mp_mod(DP,D,temp);
    mp_move(temp,Q);
    mp_dec(temp);
    mp_mod(DQ,D,temp);

    i = mp_modexp_crt(temp, inbuf, P, Q, DP, DQ, U);
    mp_burn(DP);
    mp_burn(DQ);
    if (i < 0) {
        mp_burn(temp);
        XRETURN (-1);
    }
    mp_convert_order((byte *)temp);
    front = (byte *)temp;           /* Start of block */
    i = units2bytes(global_precision);
    back = (byte *)front + i;       /* End of block */
    blocksize = countbytes(N) - 1;
    i -= blocksize;             /* Expected # of leading 0's */

    if (i < 0)              /* This shouldn't happen */
        goto Corrupted;
    while (i--)             /* Extra bytes should be 0 */
        if (*front++)
            goto Corrupted;

    /* How to distinguish old PGP from PKCS formats.
     * PGP packets have a trailing type byte (CK_ENCRYPTED_BYTE),
     * while PKCS formats have it leading.
     */
    if (front[0] != CK_ENCRYPTED_BYTE && back[-1] == CK_ENCRYPTED_BYTE) {
        /* PGP 2.0 format  - padding at the end */
        if (back[-1] != CK_ENCRYPTED_BYTE)
            goto Corrupted;
        while (*--back) /* Skip non-zero random padding */
            ;
    } else {
        /* PKCS format - padding at the beginning */
        if (*front++ != CK_ENCRYPTED_BYTE)
            goto Corrupted;
        while (*front++)    /* Skip non-zero random padding */
            ;
    }
    if (back <= front)
        goto Corrupted;
    blocksize = back-front;

    memcpy(outbuf, front, blocksize);
    mp_burn(temp);
    XRETURN (blocksize);

Corrupted:
    mp_burn(temp);
    XRETURN (-9);

#undef XRETURN
    exit_rsa_private_decrypt:
    XDEL(temp);
    return Xretval;
} /* rsa_private_decrypt */
